#
# TclTk windowing routines for cnet
#
# Written by Chris McDonald (chris@cs.uwa.edu.au)
# with thanks to Michael J. Robins.
#
set buttonfont		"-adobe-helvetica-medium-r-*-*-10-*-*-*-*-*-*-*"
set labelfont		"-adobe-helvetica-bold-r-*-*-10-*-*-*-*-*-*-*"
set stdiofont		"-*-courier-medium-r-*-*-10-*-*-*-*-*-*-*"
set cneticon		/net/misc/chris/cnetlib/icon.xbm
set stdiorows	 	24
set stdiocols	 	80
set stdiohistory 	64
set stdiobg	 	cyan
set stdiofg	 	black

proc SetWindowTitle {w str} {
    toplevel $w
    wm title $w "cnet: $str"
}

proc InitMainWindow {w topology} {
    global cneticon
    wm withdraw .
    SetWindowTitle $w $topology
    wm resizable $w 0 0
    wm iconbitmap $w "@$cneticon"
}

proc InitTitleBar {w progname deldstr exitproc} {
    global buttonfont labelfont
    global CN_DELIVERY

    frame $w
    button $w.exit	-text	"Exit $progname" \
	-font $buttonfont \
	-command $exitproc \
	-relief raised -bd 2
    label $w.msgs	\
	-font $labelfont \
	-text "Messages delivered correctly:"
    set CN_DELIVERY	"$deldstr"
    label $w.deld \
	-font $buttonfont \
	-textvariable CN_DELIVERY

    pack $w.exit $w.msgs $w.deld -side left -fill x -expand yes
    pack $w -side top -padx 2 -pady 2 -fill x -expand no
}

proc InitTopFrame {w runproc saveproc} {
    global buttonfont
    global tk_version

    set f [frame $w.top]
    button $f.run	-textvariable CN_RUN_STRING \
	-font $buttonfont \
	-command	"$runproc run" \
	-width 8 \
	-relief raised -bd 2
    button $f.single	-text	"Single event" \
	-font $buttonfont \
	-command	"$runproc step" \
	-relief raised -bd 2
    button $f.save	-text	"Save topology" \
	-font $buttonfont \
	-command	"ToggleSave $saveproc" \
	-relief raised -bd 2

    menubutton $f.subwins \
	-text Subwindows \
	-indicatoron true \
	-direction flush \
	-font $buttonfont \
	-relief raised -bd 2

    set m [menu $f.subwins.popup]
    if {$tk_version > 3.6} \
	then {$m configure -tearoff false}
    $m add command \
	-label "default node attributes" \
	-font $buttonfont \
	-command "ToggleNodeAttributes default"
    $m add command \
	-label "default link attributes" \
	-font $buttonfont \
	-command "ToggleLinkAttributes -1 -1 default $w 0 0 0 0"
    $m add command \
	-label "histograms" \
	-font $buttonfont \
	-command "ToggleHistogram"
    $m add command \
	-label "statistics" \
	-font $buttonfont \
	-command "ToggleStatistics"

    $f.subwins configure \
	-menu $m

    pack $f.run $f.single $f.save $f.subwins -side left -fill x -expand yes
    pack $f -side top -padx 4 -pady 2 -fill x -expand no
}

proc InitCanvas {w width height} {
    if {[info command $w] == ""} {
	canvas $w -width $width -height $height  \
	    -background white -relief sunken -border 2
	pack $w -side top -fill both -expand yes
     }
}

# ----------------------------------------------------------------

proc ToggleHistogram {} {
    global CN_HISTO_WIDTH CN_HISTO_HEIGHT histogram_displayed

    set w .hist
    if {[info command $w] == ""} {
	SetWindowTitle $w "histograms"
	InitCanvasFrame $w.top "histograms" "ToggleHistogram"
	InitCanvas $w.can $CN_HISTO_WIDTH $CN_HISTO_HEIGHT
	set histogram_displayed true
	histogram_created $w.can
	wm resizable $w 0 0

    } elseif {[wm state $w] == "withdrawn"} {
	wm deiconify $w
	set histogram_displayed true
    } else {
	wm withdraw $w
	set histogram_displayed false
    }
}

proc SaveTopology {saveproc e} {
    $saveproc [$e get]
    wm withdraw .save
}

proc ToggleSave {saveproc} {
    global buttonfont labelfont

    set w .save
    if {[info command $w] == ""} {
	SetWindowTitle $w "save topology"
	set s [frame $w.f]
	label $s.l \
	    -font $labelfont \
	    -text "Topology filename:"
	entry $s.e	-background white -width 30
	bind $s.e <Return> "SaveTopology $saveproc $s.e"
	pack $s.l $s.e -side left -fill both -expand yes
	pack $s

	button $w.save \
	    -font $buttonfont \
	    -text save \
	    -command "SaveTopology $saveproc $s.e"
	button $w.cancel \
	    -font $buttonfont \
	    -text cancel \
	    -command "wm withdraw $w"
	pack $w.save $w.cancel -side left -fill both -expand yes
    } elseif {[wm state $w] == "withdrawn"} {
	wm deiconify $w
    } else {
	wm withdraw $w
    }
}

proc ToggleStatistics {} {
    global CN_STATS_WIDTH CN_STATS_HEIGHT stats_displayed

    set w .stats
    if {[info command $w] == ""} {
	SetWindowTitle $w "statistics"
	InitCanvasFrame $w.top "statistics" "ToggleStatistics"
	InitCanvas $w.can $CN_STATS_WIDTH $CN_STATS_HEIGHT
	set stats_displayed true
	stats_created $w.can
	wm resizable $w 0 0

    } elseif {[wm state $w] == "withdrawn"} {
	wm deiconify $w
	set stats_displayed true
    } else {
	wm withdraw $w
	set stats_displayed false
    }
}

proc InitCanvasFrame {w name cmd} {
    global buttonfont labelfont
    global CN_UPDATE_FREQ

    frame $w
    button $w.close \
	-font	$buttonfont \
	-text	 "Close $name" \
	-command $cmd \
	-relief raised -bd 2
    label $w.update \
	-font $labelfont \
	-text $CN_UPDATE_FREQ

    pack $w.close $w.update -side left -fill x -expand yes
    pack $w -side top -fill x -expand no
}

# ----------------------------------------------------------------

proc InitNode {n nodename nodetype mainwin x y width height_stats} {

    global buttonfont labelfont
    global trace_events
    global node_choice_title node_choice_label node_choice_set

    set w .node$n
    if {$nodename == "default"} {
	set wintitle "default node attributes"
    } else {
	set wintitle "$nodename"
    }
    SetWindowTitle $w "$wintitle"

    set f [frame $w.top]
    button $f.close \
	-font	$buttonfont \
	-text	"Close $wintitle" \
	-command	"wm withdraw $w" \
	-relief raised -bd 2
    checkbutton $f.trace -text	"trace all events" \
	-font	$buttonfont \
	-variable "trace_events($n)" \
	-relief flat

    pack $f.close $f.trace -side left -anchor e -expand no
    pack $w.top -side top -fill x -expand no

    if {$nodetype != "router"} {
	set nn 0
	if {$nodename == "default"} {
	    set nn -1
	}
	foreach d {0 1 2} {
	    set f [frame $w.choice$d]
	    label $f.lab \
		-width 20 \
		-font	$labelfont \
		-text	$node_choice_title($d)
	    pack $f.lab -anchor w -side left
	    foreach c {0 1 2 3} {
		radiobutton $f.$c \
		    -width  9 \
		    -font	$buttonfont \
		    -text   $node_choice_label($nn,$d,$c) \
		    -value $c \
		    -variable $f.selected \
		    -relief raised -bd 2 \
		    -command "node_choice $n $d $c"
	    }
	    $f.$node_choice_set($nn,$d) select
	    pack $f.0 $f.1 $f.2 $f.3 -side left -fill x -expand yes
	}
	pack $w.choice0 $w.choice1 $w.choice2 -side top -fill x -expand no
    }
    if {$nodetype != "default"} {
	if {$nodetype == "host"} {
	    InitCanvas $w.stats $width $height_stats
	}

	set f [frame $w.d]
	foreach i {0 1 2 3 4} {
	    button "$f.debug$i"	-text " " \
		-font	$buttonfont \
		-command	"node_debug_button $n $i" \
		-width 8 \
		-relief raised -bd 2
	}
	pack $f.debug0 $f.debug1 $f.debug2 $f.debug3 $f.debug4 \
	    -side left -fill x -expand yes
	pack $f -side top -padx 4 -pady 2 -fill x -expand no

	stdioinit $w $n
	node_created $n $w $w.stats $w.stdio
    }
    set node_displayed($n) false
    wm withdraw $w
    wm resizable $w 0 0
}

proc ToggleNodeAttributes {n} {
    global node_displayed

    set w .node$n
    if {[wm state $w] == "withdrawn"} {
	wm deiconify $w
	set node_displayed($n) true
    } else {
	wm withdraw $w
	set node_displayed($n) false
    }
}

proc SetDebugString {w str} {
    global buttonfont
    $w configure \
	-font $buttonfont \
	-text $str
}

# ----------------------------------------------------------------

proc ToggleLinkAttributes {from to linkname mainwin x y \
					stats_width stats_height} {
    global buttonfont labelfont
    global link_scale_title link_scale_max link_scale_value
    global link_displayed

    set w .link($from,$to)
    if {[info command $w] == ""} {

	regsub -all "(\\+|\\x)" [wm geometry $mainwin] " " geom
	if {$linkname == "default"} {
	    set x [expr [lindex $geom 0] + [lindex $geom 2] - 100]
	    set y [expr [lindex $geom 1] + [lindex $geom 3] +  20]
	    set wintitle "default link attributes"
	} else {
	    set x [expr $x + [lindex $geom 2] + 30]
	    set y [expr $y + [lindex $geom 3] + 90]
	    set wintitle "$linkname link"
	}
	SetWindowTitle $w "$wintitle"
	wm geometry $w +$x+$y

	set t [frame $w.top]
	button $t.close \
	    -font $buttonfont \
	    -text	"Close $linkname link attributes" \
	    -command	"wm withdraw $w" \
	    -relief raised -bd 2
	pack $t.close -side left -anchor e \
	    -expand no
	pack $t -side top -fill x -expand yes

	set min  0
	if {$linkname != "default"} {
	    InitCanvas $w.stats $stats_width $stats_height
	    set min -1
	}

	link_created $w $from $to $w.stats

	set b [frame $w.bottom]
	foreach c {0 1 2 3} {
	    set f [frame $b.scale$c]
	    label $f.lab \
		-width 25 \
		-font	$labelfont \
		-text	$link_scale_title($c)
	    pack $f.lab -anchor w -side left
	    scale $f.s	-from $min -to $link_scale_max($c) \
			-orient horizontal \
			-length 100 \
			-showvalue 1 \
			-command "link_scale $from $to $c"
	    $f.s set $link_scale_value($c)
	    pack $f.s -side left -fill x -expand yes
	}

	pack $b.scale0 $b.scale1 $b.scale2 $b.scale3 \
	    -side top -fill x -expand no
	pack $b -side top -fill x -expand yes

	set link_displayed($from,$to) true
	wm resizable $w 0 0

    } elseif {[wm state $w] == "withdrawn"} {
	wm deiconify $w
	set link_displayed($from,$to) true
    } else {
	wm withdraw $w
	set link_displayed($from,$to) false
    }
}

# ----------------------------------------------------------------

set node_menu_displayed -1
# only allow one nodemenu

proc PopupNodeMenu {n nodename mainwin activemask x y} {
    global node_menu_label node_menu_displayed tk_version

    set m .nodemenu
    if {[info command $m] == ""} {
	if {$tk_version > 3.6} \
	    then {menu $m -tearoff false} \
	    else {menu $m}

	$m add command -label $nodename -state disabled
	$m add separator
	foreach s {0 1 2 3 4} {
	    $m add command -label $node_menu_label($s) \
			   -command "PopdownNodeMenu $m $s"
	}
    }
    if {$node_menu_displayed != -1} {
	set node_menu_displayed -1
	$m unpost
	return
    }

    $m entryconfigure 0 -label $nodename
    foreach s {0 1 2 3 4} {
	$m entryconfigure [expr $s + 2] \
	    -state [expr {[expr $activemask&(1<<$s)] ? "normal" : "disabled"}]
    }
    set node_menu_displayed $n

    regsub -all "(\\+|\\x)" [wm geometry $mainwin] " " geom
    set x [expr $x + [lindex $geom 2]]
    set y [expr $y + [lindex $geom 3] + 10]
    $m post $x $y
}

proc PopdownNodeMenu {m s} {
    global node_menu_displayed node_menu_select

    node_menu_select $node_menu_displayed $s
    set node_menu_displayed -1
    $m unpost
}


set link_menu_displayed -1
# only allow one linkmenu

proc PopupLinkMenu {l linkname mainwin activemask x y} {
    global link_menu_label link_menu_displayed tk_version

    set m .linkmenu
    if {[info command $m] == ""} {
	if {$tk_version > 3.6} \
	    then {menu $m -tearoff false} \
	    else {menu $m}

	$m add command -label $linkname -state disabled
	$m add separator
	foreach s {0 1 2 3} {
	    $m add command -label $link_menu_label($s) \
			   -command "PopdownLinkMenu $m $s"
	}
    }
    if {$link_menu_displayed != -1} {
	set link_menu_displayed -1
	$m unpost
	return
    }

    $m entryconfigure 0 -label $linkname
    foreach s {0 1 2 3} {
	$m entryconfigure [expr $s + 2] \
	    -state [expr {[expr $activemask&(1<<$s)] ? "normal" : "disabled"}]
    }
    set link_menu_displayed $l

    regsub -all "(\\+|\\x)" [wm geometry $mainwin] " " geom
    set x [expr $x + [lindex $geom 2]]
    set y [expr $y + [lindex $geom 3] + 10]
    $m post $x $y
}

proc PopdownLinkMenu {m s} {
    global link_menu_displayed link_menu_select

    link_menu_select $link_menu_displayed $s
    set link_menu_displayed -1
    $m unpost
}

# ------------------------------------------------------------------------

proc stdioinit {w n} {
    global stdiowin stdiofont stdiolc stdioinput
    global stdiorows stdiocols stdiobg stdiofg

    set win		$w.stdio
    set stdiowin($n)	$win
    set stdiolc($n)	1
    set stdioinput($n)	""

    text $win \
	-width $stdiocols \
	-height $stdiorows \
	-font $stdiofont \
	-background $stdiobg \
	-foreground $stdiofg \
	-wrap none \
	-yscrollcommand "$w.ysbar set"
    scrollbar $w.ysbar -orient vertical   -command "$win yview"
    pack $win $w.ysbar -side left -fill both -expand no

    bind $win <Enter>    { focus %W ; break }
    bind $win <KeyPress> "stdiokey $n %K %A ; break"
}

proc stdioscroll {n} {
    global stdiowin stdiohistory

    set win $stdiowin($n)
    for {set l1 1; set l2 2} {$l1 < $stdiohistory} {incr l1; incr l2} {
	$win delete $l1.0 $l1.999
	$win insert $l1.0 [$win get $l2.0 $l2.999]
    }
    $win delete $stdiohistory.0 $stdiohistory.999
}

proc stdiokey {n keysym ascii} {
    global stdiowin stdiolc stdioinput stdiohistory

    set win $stdiowin($n)
    if {$keysym == "BackSpace" || $keysym == "Delete"} then {
	set len [string length $stdioinput($n)]
	if {$len == 0} then {
	    bell
	} else {
	    set s $stdioinput($n)
	    set s [string range $s 0 [expr $len - 2]]
	    set stdioinput($n) $s
	    $win delete "insert - 1 char"
	}
    } elseif {$keysym == "Return" } then {
	if {$stdiolc($n) < $stdiohistory} then {
	    $win insert $stdiolc($n).end "\n"
	    incr stdiolc($n)
	} else {
	    stdioscroll $n
	}
	stdio_input $n "$stdioinput($n)"
	set stdioinput($n)	""
    } else {
	set stdioinput($n) "$stdioinput($n)$ascii"
	$win insert $stdiolc($n).end $ascii
    }
}

proc stdioclr {n} {
    global stdiowin stdiolc

    set win $stdiowin($n)
    $win delete 1.0 end
    set stdiolc($n) 1
}

proc stdioaddstr {n str newline} {
    global stdiowin stdiohistory stdiolc

    set win $stdiowin($n)
    if {$stdiolc($n) < $stdiohistory} then {
	if {$newline == "true"} then {
	    $win insert $stdiolc($n).end "$str\n"
	    incr stdiolc($n)
	} else {
	    $win insert $stdiolc($n).end "$str"
	}
    } else {
	$win insert $stdiohistory.end "$str"
	if {$newline == "true"} then {
	    stdioscroll $n
	}
    }
}
